"""Function for reading Data Package properties."""

import json
from pathlib import Path
from typing import Any
from urllib import parse, request
from urllib.error import HTTPError, URLError

from check_datapackage import check

from seedcase_flower.errors import (
    FileDoesNotExistError,
    HTTPDomainError,
    HTTPStatusError,
    JSONFormatError,
    NotJSONError,
)
from seedcase_flower.parse_source import Address

JSON_CONTENT_TYPES = ("application/json", "application/ld+json", "application/geo+json")


def read_properties(address: Address) -> dict[str, Any]:
    """Read properties from a local or remote datapackage."""
    datapackage: dict[str, Any]
    if address.local:
        path = Path(parse.urlsplit(address.value).path)
        try:
            with open(path) as properties_file:
                datapackage = json.load(properties_file)
        except FileNotFoundError:
            raise FileDoesNotExistError(str(path))
        except json.JSONDecodeError as e:
            raise JSONFormatError(str(path), str(e))
    else:
        try:
            with request.urlopen(address.value) as open_url:  # nosec B310
                content_type = open_url.headers.get("Content-Type", "")
                if not content_type.startswith(JSON_CONTENT_TYPES + ("text/plain",)):
                    main_type = content_type.split(";")[0].strip()
                    raise NotJSONError(address.value, main_type)
                datapackage = json.load(open_url)
        except HTTPError as e:
            raise HTTPStatusError(address.value, e.code, e.reason)
        except URLError as e:
            if "Name or service not known" in str(
                e.reason
            ) or "getaddrinfo failed" in str(e.reason):
                raise HTTPDomainError(address.value)
            raise JSONFormatError(address.value, str(e))
        except json.JSONDecodeError as e:
            raise JSONFormatError(address.value, str(e))
    check(datapackage, error=True)
    return datapackage



from pathlib import Path

from pytest import raises

from seedcase_flower.build_sections import BuiltSection
from seedcase_flower.write_sections import write_sections


def test_writes_sections_to_file(tmp_path):
    built_sections = [
        BuiltSection(
            content="Section 1 content.",
            output_path=Path("section1.qmd"),
        ),
        BuiltSection(
            content="Section 2 content.",
            output_path=Path("some-folder") / "section2.qmd",
        ),
    ]

    output_files = write_sections(built_sections, tmp_path)

    assert output_files == [
        tmp_path / "section1.qmd",
        tmp_path / "some-folder" / "section2.qmd",
    ]
    assert output_files[0].read_text() == "Section 1 content."
    assert output_files[1].read_text() == "Section 2 content."


def test_throws_error_if_no_output_path(tmp_path):
    built_sections = [
        BuiltSection(
            content="Section 1 content.",
        ),
        BuiltSection(
            content="Section 2 content.",
            output_path=Path("section2.qmd"),
        ),
    ]

    with raises(ValueError):
        write_sections(built_sections, tmp_path)

    assert not (tmp_path / "section2.qmd").exists()
    assert not (tmp_path / "section1.qmd").exists()



"""Build sections from templates."""

import tomllib
from dataclasses import dataclass
from importlib.resources import files
from pathlib import Path
from typing import Any, Optional, Union, cast

from jinja2 import Environment, FileSystemLoader, Template, select_autoescape
from jsonpath import findall, finditer
from pydantic import BaseModel, Field

from seedcase_flower.config import Config
from seedcase_flower.internals import _flat_map, _map
from seedcase_flower.sections import Content, Many, ManyContent, One
from seedcase_flower.styles import Style


class SectionsToml(BaseModel, frozen=True):
    """Data model of the contents of the `sections.toml` file.

    Attributes:
        one_sections: Sections in the config file generating one file.
        many_sections: Sections in the config file generating many files.
    """

    one_sections: list[One] = Field(alias="one", default=[])
    many_sections: list[Many] = Field(alias="many", default=[])


@dataclass(frozen=True)
class BuiltSection:
    """A section that has been built from a template.

    Attributes:
        content: The rendered template as a string.
        output_path: The path where the section should be written, relative
            to `Config.output_dir`. Sections displayed in the terminal don't
            have an `output_path`.
    """

    content: str
    output_path: Optional[Path] = None


def _get_template_dir(style: Style) -> Path:
    styles_path = Path(str(files("seedcase_flower").joinpath("styles")))
    return styles_path / style.name


def _load_sections_toml(template_dir: Path) -> SectionsToml:
    if not template_dir.is_dir():
        raise NotADirectoryError(f"Template directory '{template_dir}' does not exist.")

    toml_path = template_dir / "sections.toml"
    if not toml_path.is_file():
        raise FileNotFoundError(
            f"Template directory '{template_dir}' does not contain a sections.toml "
            "file."
        )

    with open(toml_path, mode="rb") as file:
        toml_file = tomllib.load(file)

    return SectionsToml.model_validate(toml_file)


def _inline_code(value: Optional[str]) -> Optional[str]:
    return f"`{value}`" if value else None


def _inline_code_list(value: Union[str, list[str]]) -> str:
    # Some Data Package fields allow either a string or a list of strings
    if isinstance(value, str):
        value = [value]
    return ", ".join(_map(value, lambda item: f"`{item}`"))


def _max_column_width(rows: list[list[str]], col: int) -> int:
    return max(map(lambda row: len(row[col]), rows), default=0)


def _cell_width(headers: list[str], rows: list[list[str]], col: int) -> int:
    return max(len(headers[col]), _max_column_width(rows, col))


def _column_widths(headers: list[str], rows: list[list[str]]) -> list[int]:
    return _map(range(len(headers)), lambda i: _cell_width(headers, rows, i))


def _format_row(row: list[str], widths: list[int]) -> str:
    return "| " + " | ".join(map(str.ljust, row, widths)) + " |"


def _separator_row(widths: list[int]) -> str:
    return "|" + "|".join(map(lambda w: "-" * (w + 2), widths)) + "|"


def _render_markdown_table(headers: list[str], rows: list[list[str]]) -> str:
    if not headers:
        return ""

    widths = _column_widths(headers, rows)

    return "\n".join(
        [
            _format_row(headers, widths),
            _separator_row(widths),
            *map(lambda row: _format_row(row, widths), rows),
        ]
    )


def _create_jinja_env(template_dir: Path) -> Environment:
    env = Environment(
        loader=FileSystemLoader(template_dir),
        trim_blocks=True,
        autoescape=select_autoescape(
            enabled_extensions=(
                "html.jinja",
                "xml.jinja",
            ),
            default=False,
        ),
    )
    # Render a list of strings as comma-separated inline code
    env.filters["_inline_code_list"] = _inline_code_list
    # Render a single value as inline code
    env.filters["_inline_code"] = _inline_code
    # Render a markdown table with adjusted column widths
    env.globals["_render_markdown_table"] = _render_markdown_table
    return env


def _get_template(
    template_dir: Path, content_template_path: Path, env: Environment
) -> Template:
    template_path = template_dir / content_template_path
    if not template_path.is_file():
        raise FileNotFoundError(
            f"Template file '{template_path}' does not exist in the template directory."
        )
    return env.get_template(template_path.name)


def _build_content_one(
    content: Content, properties: dict[str, Any], template_dir: Path, env: Environment
) -> str:
    selected_properties = findall(content.jsonpath, properties)
    if len(selected_properties) > 1:
        raise ValueError(
            "A `one` section expects at most one match. JSON path "
            f"{content.jsonpath!r} returned {len(selected_properties)} "
            "matches. Use a more specific JSON path or switch to a `many` section."
        )

    template = _get_template(template_dir, content.template_path, env)
    return template.render(
        **{
            content.jinja_variable: selected_properties[0]
            if selected_properties
            else None
        }
    )


def _build_one(
    one: One, properties: dict[str, Any], template_dir: Path, env: Environment
) -> BuiltSection:
    built_contents = _map(
        one.contents,
        lambda content: _build_content_one(content, properties, template_dir, env),
    )
    return BuiltSection(
        output_path=one.output_path,
        content="\n".join(built_contents),
    )


@dataclass(frozen=True)
class ManyMatch:
    """A metadata item displayed in a `many` section.

    Attributes:
        resource_name: The name of the resource containing the metadata item.
        properties: The metadata item.
    """

    resource_name: str
    properties: dict[str, Any]


def _get_many_matches(jsonpath: str, properties: dict[str, Any]) -> list[ManyMatch]:
    return _map(
        finditer(jsonpath, properties),
        lambda match: ManyMatch(
            resource_name=properties["resources"][match.parts[1]]["name"],
            properties=cast(dict[str, Any], match.value),
        ),
    )


def _build_many(
    many: Many, properties: dict[str, Any], template_dir: Path, env: Environment
) -> list[BuiltSection]:
    template = _get_template(template_dir, many.template_path, env)
    matches = _get_many_matches(many.content.jsonpath, properties)

    return _map(
        matches,
        lambda match: BuiltSection(
            output_path=_get_output_path_for_match(match, many),
            content=template.render(**{many.jinja_variable: match.properties}),
        ),
    )


def _get_output_path_for_match(match: ManyMatch, many: Many) -> Optional[Path]:
    if not many.output_path:
        return None

    return Path(
        many.output_path.as_posix()
        .replace(ManyContent.resources.placeholder, match.resource_name)
        .replace(ManyContent.fields.placeholder, match.properties["name"])
    )


def build_sections(properties: dict[str, Any], config: Config) -> list[BuiltSection]:
    """Builds the output files based on the provided properties and configuration.

    Args:
        properties: The Data Package properties.
        config: The configuration.

    Returns:
        A list of built output files.
    """
    template_dir = config.template_dir or _get_template_dir(config.style)
    sections_toml = _load_sections_toml(template_dir)
    env = _create_jinja_env(template_dir)
    return _map(
        sections_toml.one_sections,
        lambda one: _build_one(one, properties, template_dir, env),
    ) + _flat_map(
        sections_toml.many_sections,
        lambda many: _build_many(many, properties, template_dir, env),
    )



import re
from enum import Enum
from pathlib import Path
from typing import Annotated, Optional, Self

from jsonpath import JSONPathSyntaxError, compile
from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    field_validator,
    model_validator,
)


def _check_path_relative(value: Path) -> Path:
    if value.is_absolute():
        raise ValueError(
            f"{value!r} is an absolute path. Please provide a path relative to the "
            "parent folder."
        )
    return value


type RelativePath = Annotated[Path, AfterValidator(_check_path_relative)]


def _check_template_extension(value: Path) -> Path:
    if len(value.suffixes) < 2 or value.suffix != ".jinja":
        raise ValueError(
            f"The template name '{value.name}' must be of the form "
            "`<name>.<ext>.jinja`."
        )
    return value


type TemplatePath = Annotated[RelativePath, AfterValidator(_check_template_extension)]


class KebabModel(BaseModel, frozen=True):
    """Allow creating Pydantic model from kebab-case data."""

    model_config = ConfigDict(
        alias_generator=lambda string: string.replace("_", "-"),
        populate_by_name=True,
    )


class Content(KebabModel, frozen=True):
    """Content to include within a `One` section.

    The `Content` class defines what Data Package properties and
    [Jinja2](https://jinja.palletsprojects.com/en/stable/) template file belong within
    a specific section (an output file) in the documentation. You can use
    this class to customise how different parts of the `datapackage.json` file
    are displayed in the documentation and to create common presets when
    styles share similar content structures.

    Attributes:
        jsonpath: The JSON path, expressed using
            [JSON path syntax](https://jg-rp.github.io/python-jsonpath/syntax/),
            to the Data Package property that should be sent to the `template_path`
            Jinja2 file.
        template_path: The path to the Jinja2 template file for this content
            item, relative to `Config.template_dir`.
        jinja_variable: The Jinja2 variable name that will be used in the template
            to reference this content item.

    Examples:
        ```{python}
        import seedcase_flower as fl
        from pathlib import Path

        # A content item displaying the metadata of the whole package
        package_content = fl.Content(
            jsonpath="$",
            template_path=Path("package.qmd.jinja"),
            jinja_variable="package",
        )
        ```
    """

    jsonpath: str
    template_path: TemplatePath
    jinja_variable: str

    @field_validator("jsonpath", mode="after")
    @classmethod
    def _check_jsonpath(cls, value: str) -> str:
        try:
            compile(value)
        except JSONPathSyntaxError:
            raise ValueError(
                f"{value!r} is not a correct JSON path. See "
                "https://jg-rp.github.io/python-jsonpath/syntax/ for the expected "
                "syntax."
            )
        return value


class One(KebabModel, frozen=True):
    """A section of the documentation that outputs to one file.

    See the [design](https://flower.seedcase-project.org/docs/design/interface/config#section)
    for an explanation of the design of `One`.
    See the [guide](https://flower.seedcase-project.org/docs/guide/custom-styles) on
    how to set up custom styles and sections.
    See `Config` for more details on the top-level settings and `Content` for
    more details on the content items.

    Attributes:
        output_path: The output path for the section relative to
            `Config.output_dir`. Can be `None` when a style outputs to the terminal
            such as when using `view()`.
        contents: List of content items to include in this
            section. See `Content` for more details about what to include. If more than
            one content item is included, they will be concatenated in the order
            provided, so that the `output_path` file will contain the
            output of the rendered Jinja2 templates for each content item, appended one
            after the other.

    Examples:
        ```{python}
        import seedcase_flower as fl
        from pathlib import Path

        # A section that contains only the package and contributors content items,
        # saved to the `package.qmd` file.
        section = fl.One(
            output_path=Path("package.qmd"),
            contents=[
                fl.Content(
                    jsonpath="$",
                    template_path=Path("package.qmd.jinja"),
                    jinja_variable="package",
                ),
                fl.Content(
                    jsonpath="$.contributors",
                    template_path=Path("contributors.qmd.jinja"),
                    jinja_variable="contributors",
                ),
            ],
        )
        ```
    """

    output_path: Optional[RelativePath] = None
    contents: list[Content]


class ManyContent(str, Enum):
    """Content item to include within a `Many` section."""

    resources = "resources"
    fields = "fields"

    @property
    def jsonpath(self) -> str:
        """The JSON path that selects this content item."""
        return {
            self.resources: "$.resources[*]",
            self.fields: "$.resources[*].schema.fields[*]",
        }[self]

    @property
    def placeholder(self) -> str:
        """The placeholder for the name of this content item in the output path."""
        return {
            self.resources: "{resource-name}",
            self.fields: "{field-name}",
        }[self]

    @property
    def allowed_placeholders(self) -> list[list[str]]:
        """The allowed placeholders in the output path for this content item."""
        resource = ManyContent.resources.placeholder
        field = ManyContent.fields.placeholder
        return {
            self.resources: [[resource]],
            self.fields: [[resource], [field], [resource, field]],
        }[self]


class Many(KebabModel, frozen=True):
    """A section of the documentation that outputs to multiple files.

    See the [design](https://flower.seedcase-project.org/docs/design/interface/config#section)
    for an explanation of the design of `Many`.
    See the [guide](https://flower.seedcase-project.org/docs/guide/custom-styles) on
    how to set up custom styles and sections.
    See `Config` for more details on the top-level settings and `Content` for
    more details on the content items.

    Attributes:
        output_path: The output path for the section relative to `Config.output_dir`.
            If a directory is provided, the output files will be saved in that
            directory. If the path includes a placeholder in curly brackets (e.g.
            `resources/{resource-name}.qmd`), the output files will be named following
            this template.
        content: The content item to display in this section; choices are `resources`
            and `fields`.

    Examples:
        ```{python}
        import seedcase_flower as fl
        from pathlib import Path

        # A section that displays each resource in a separate file within the
        # `resources/` folder.
        section = fl.Many(
            output_path=Path("resources/"),
            content=fl.ManyContent.resources,
            template_path=Path("resource.md.jinja"),
            jinja_variable="resource",
        )
        ```
    """

    output_path: Optional[RelativePath] = None
    content: ManyContent
    template_path: TemplatePath
    jinja_variable: str

    @property
    def extension(self) -> str:
        """The extension of the files to create."""
        return "".join(self.template_path.suffixes[:-1])

    @model_validator(mode="after")
    def _check_extensions(self) -> Self:
        if not self.output_path or not self.output_path.suffixes:
            return self
        if "".join(self.output_path.suffixes) != self.extension:
            raise ValueError(
                f"The output path '{self.output_path}' must have the same file "
                "extension as the files generated from the "
                f"template '{self.template_path}'."
            )
        return self

    @model_validator(mode="after")
    def _check_placeholders(self) -> Self:
        if not self.output_path:
            return self

        posix_path = self.output_path.as_posix()
        placeholders = re.findall(r"\{[^\}]*\}", posix_path)
        if placeholders and placeholders not in self.content.allowed_placeholders:
            # e.g. `resources/{not-resource-name}.qmd`
            raise ValueError(
                f"The output path '{self.output_path}' contains incorrect "
                "placeholders. Check that you're using the correct placeholders "
                # TODO: Add link
                "for the content type in the correct order. See LINK for more details."
            )

        if not self.output_path.suffix:
            # Normalise folder paths
            # e.g. `resources/` -> `resources/{resource-name}.qmd`
            object.__setattr__(
                self,
                "output_path",
                self.output_path / (self.content.placeholder + self.extension),
            )
            return self

        if not placeholders:
            # e.g. `resources/concrete-file.qmd`
            raise ValueError(
                f"The output path '{self.output_path}' points to a file. Output "
                "paths in `many` sections must point to a folder or use a "
                "placeholder for the filename."
            )
        if self.content.placeholder not in placeholders:
            # e.g. `folder/{resource-name}.qmd` in a fields section
            raise ValueError(
                f"The output path '{self.output_path}' uses the wrong placeholder "
                "in the filename. Files displaying fields should use the placeholder "
                f"{ManyContent.fields.placeholder!r} in the filename."
            )

        return self



"""Module containing all source code."""

from . import errors
from .cli import build, view
from .config import Config
from .sections import Content, Many, ManyContent, One
from .styles import Style, ViewStyle

__all__ = [
    "build",
    "view",
    "Config",
    "Style",
    "ViewStyle",
    "Content",
    "One",
    "Many",
    "ManyContent",
]



"""Shared fixtures for tests."""

import json

import pytest


@pytest.fixture
def datapackage():
    """Return a valid datapackage dict with resources."""
    return {
        "name": "test-package",
        "title": "Test Package",
        "description": "A test datapackage",
        "version": "1.0.0",
        "licenses": [{"name": "MIT"}],
        "resources": [
            {
                "name": "data",
                "path": "data.csv",
                "schema": {
                    "fields": [
                        {"name": "id", "type": "integer"},
                        {"name": "name", "type": "string"},
                    ]
                },
            },
            {
                "name": "data2",
                "path": "data2.csv",
                "schema": {
                    "fields": [
                        {"name": "id", "type": "integer"},
                        {"name": "age", "type": "integer"},
                    ]
                },
            },
        ],
    }


@pytest.fixture
def datapackage_path(tmp_path, datapackage):
    """Create a temporary datapackage.json file and return its path as a string."""
    file_path = tmp_path / "datapackage.json"
    file_path.write_text(json.dumps(datapackage))
    return str(file_path)



"""Tests for parsing the source for the Data Package."""

import pytest

from seedcase_flower.parse_source import Address, parse_source

# parse_source: plain path (no scheme) ====


def test_parse_source_plain_file_path_is_local(tmp_path):
    """A plain file path with no scheme should return a local Address."""
    result = parse_source(str(tmp_path / "datapackage.json"))
    assert result.local is True


def test_parse_source_plain_file_path_has_file_scheme(tmp_path):
    """A plain file path should be normalised to a file path."""
    result = parse_source(str(tmp_path / "datapackage.json"))
    assert result.value.startswith("file://")


def test_parse_source_directory_path_appends_datapackage_json(tmp_path):
    """Passing a directory path should append datapackage.json to the source."""
    result = parse_source(str(tmp_path))
    assert result.value.endswith("datapackage.json")


def test_parse_source_directory_path_is_local(tmp_path):
    """Passing a directory path should return a local Address."""
    result = parse_source(str(tmp_path))
    assert result.local is True


# parse_source: `file:` scheme ====


def test_parse_source_file_scheme_is_local(tmp_path):
    """A file path should return a local Address."""
    result = parse_source(f"file:{tmp_path / 'datapackage.json'}")
    assert result.local is True


def test_parse_source_file_scheme_preserves_path(tmp_path):
    """A file path pointing to a file should preserve the path."""
    file = tmp_path / "datapackage.json"
    result = parse_source(f"file:{file}")
    assert str(file) in result.value


# parse_source: `https://` scheme ====


def test_parse_source_https_is_not_local():
    """An https:// should return a non-local Address."""
    result = parse_source("https://example.com/datapackage.json")
    assert result.local is False


def test_parse_source_https_preserves_url():
    """An https:// should be returned unchanged."""
    url = "https://example.com/datapackage.json"
    result = parse_source(url)
    assert result.value == url


# parse_source: `gh:` and `github:` scheme ====


@pytest.mark.parametrize("scheme", ["gh", "github"])
def test_parse_source_github_scheme_converts_to_raw_githubusercontent(scheme):
    """GitHub sources should be converted to a raw.githubusercontent.com URL."""
    result = parse_source(f"{scheme}:owner/repo")
    assert result.value.startswith("https://raw.githubusercontent.com/")


@pytest.mark.parametrize("scheme", ["gh", "github"])
def test_parse_source_github_scheme_is_not_local(scheme):
    """GitHub sources should return a non-local Address."""
    result = parse_source(f"{scheme}:owner/repo")
    assert result.local is False


@pytest.mark.parametrize("scheme", ["gh", "github"])
def test_parse_source_github_scheme_appends_datapackage_json(scheme):
    """GitHub sources should point to the datapackage.json on the main branch."""
    result = parse_source(f"{scheme}:owner/repo")
    assert result.value.endswith("datapackage.json")


@pytest.mark.parametrize("scheme", ["gh", "github"])
def test_parse_source_github_scheme_uses_main_by_default(scheme):
    """GitHub sources without @ref should use main."""
    result = parse_source(f"{scheme}:owner/repo")
    assert result.value == (
        "https://raw.githubusercontent.com/owner/repo/main/datapackage.json"
    )


@pytest.mark.parametrize("scheme", ["gh", "github"])
def test_parse_source_github_scheme_with_tag(scheme):
    """GitHub sources with @tag should use the tag as the ref."""
    result = parse_source(f"{scheme}:owner/repo@v1.0.0")
    assert result.value == (
        "https://raw.githubusercontent.com/owner/repo/v1.0.0/datapackage.json"
    )


@pytest.mark.parametrize("scheme", ["gh", "github"])
def test_parse_source_github_scheme_with_branch(scheme):
    """GitHub sources with @branch should use the branch as the ref."""
    result = parse_source(f"{scheme}:owner/repo@develop")
    assert result.value == (
        "https://raw.githubusercontent.com/owner/repo/develop/datapackage.json"
    )


# parse_source: unsupported scheme ====


def test_parse_source_unsupported_scheme_raises_value_error():
    """An unsupported source scheme should raise a ValueError."""
    with pytest.raises(ValueError, match="source must be either"):
        parse_source("ftp:example.com/datapackage.json")


def test_parse_source_returns_address_instance(tmp_path):
    """parse_source should always return a address instance."""
    result = parse_source(str(tmp_path / "datapackage.json"))
    assert isinstance(result, Address)


# parse_source: GitHub validation errors ====


@pytest.mark.parametrize("scheme", ["gh", "github"])
def test_parse_source_github_missing_repo_raises_error(scheme):
    """GitHub source without repo should raise ValueError."""
    with pytest.raises(ValueError, match="must be in format"):
        parse_source(f"{scheme}:owner")


@pytest.mark.parametrize("scheme", ["gh", "github"])
def test_parse_source_github_missing_owner_raises_error(scheme):
    """GitHub source without owner should raise ValueError."""
    with pytest.raises(ValueError, match="must be in format"):
        parse_source(f"{scheme}:/repo")


@pytest.mark.parametrize("scheme", ["gh", "github"])
def test_parse_source_github_empty_ref_raises_error(scheme):
    """GitHub source with empty @ref should raise ValueError."""
    with pytest.raises(ValueError, match="ref after '@' cannot be empty"):
        parse_source(f"{scheme}:owner/repo@")



from pathlib import Path

from seedcase_flower.config import Config
from seedcase_flower.styles import Style


def test_creates_config():
    config = Config()
    assert config.style == Style.quarto_one_page
    assert config.template_dir is None
    assert config.output_dir == Path("docs/")

    config = Config(
        style=Style.quarto_resource_listing,
        template_dir=Path("templates/"),
        output_dir=Path("my-docs/"),
    )
    assert config.style == Style.quarto_resource_listing
    assert config.template_dir == Path("templates/")
    assert config.output_dir == Path("my-docs/")

    config = Config(template_dir=Path("templates/"))
    assert config.style == Style.quarto_one_page
    assert config.template_dir == Path("templates/")
    assert config.output_dir == Path("docs/")



from pathlib import Path
from typing import cast

from seedcase_flower.build_sections import BuiltSection
from seedcase_flower.internals import _filter, _map


def write_sections(built_sections: list[BuiltSection], output_dir: Path) -> list[Path]:
    """Writes each built section to a file in the output directory.

    Args:
        built_sections: The list of built sections to be written to files.
        output_dir: The directory where the built sections should be written.

    Returns:
        A list of paths to the output files that were created.

    Raises:
        ValueError: If any of the built sections do not have an output path.
    """
    if _filter(built_sections, lambda section: section.output_path is None):
        raise ValueError(
            "At least one section in `sections.toml` is missing an output path. "
            "When using the `build` command, all sections must have an output path."
        )

    return _map(built_sections, lambda section: _write_section(section, output_dir))


def _write_section(built_section: BuiltSection, output_dir: Path) -> Path:
    output_path = output_dir / cast(Path, built_section.output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(built_section.content)
    return output_path



from pathlib import Path
from typing import Any

from pydantic import ValidationError
from pytest import fixture, mark, raises

from seedcase_flower.build_sections import BuiltSection, build_sections
from seedcase_flower.config import Config
from seedcase_flower.internals import _flat_map, _map
from seedcase_flower.styles import Style

# One section ====

content = """
[[one.contents]]
jsonpath = "$"
template-path = "template.qmd.jinja"
jinja-variable = "package"
"""

one_content_toml = f"""
[[one]]
output-path = "section1.qmd"
{content}
"""

multi_content_toml = f"""
{one_content_toml}
{content}
"""

multi_section_toml = f"""
[[one]]
output-path = "section1.qmd"
{content}
{content}
[[one]]
output-path = "section2.qmd"
{content}
"""

properties: dict[str, Any] = {
    "name": "example-datapackage",
    "resources": [
        {
            "name": "example-resource",
            "path": "data/example-resource.csv",
        }
    ],
}


@fixture
def _template(tmp_path):
    (tmp_path / "template.qmd.jinja").write_text("{{ package.name }}")


@mark.parametrize(
    "sections_toml, expected",
    [
        (
            one_content_toml,
            [
                BuiltSection(
                    output_path=Path("section1.qmd"), content="example-datapackage"
                )
            ],
        ),
        (
            multi_content_toml,
            [
                BuiltSection(
                    output_path=Path("section1.qmd"),
                    content="example-datapackage\nexample-datapackage",
                ),
            ],
        ),
        (
            multi_section_toml,
            [
                BuiltSection(
                    output_path=Path("section1.qmd"),
                    content="example-datapackage\nexample-datapackage",
                ),
                BuiltSection(
                    output_path=Path("section2.qmd"), content="example-datapackage"
                ),
            ],
        ),
    ],
)
def test_matches_sections_toml_config(sections_toml, expected, tmp_path, _template):
    (tmp_path / "sections.toml").write_text(sections_toml)
    config = Config(template_dir=tmp_path)

    built_sections = build_sections(properties, config)

    assert built_sections == expected


def test_can_use_multiple_templates_with_different_jsonpaths(tmp_path, _template):
    sections_toml = """
    [[one]]
    output-path = "section1.qmd"

    [[one.contents]]
    jsonpath = "$"
    template-path = "template.qmd.jinja"
    jinja-variable = "package"

    [[one.contents]]
    jsonpath = "$.resources"
    template-path = "resources.qmd.jinja"
    jinja-variable = "resources"
    """
    (tmp_path / "sections.toml").write_text(sections_toml)
    (tmp_path / "resources.qmd.jinja").write_text("{{ resources | length }}")
    config = Config(template_dir=tmp_path)

    built_sections = build_sections(properties, config)

    assert built_sections == [
        BuiltSection(
            output_path=Path("section1.qmd"), content="example-datapackage\n1"
        ),
    ]


def test_uses_style_when_no_template_dir_given():
    config = Config(style=Style.quarto_one_page)

    built_sections = build_sections(properties, config)

    assert len(built_sections) == 1
    assert built_sections[0].output_path == Path("index.qmd")
    assert "example-datapackage" in built_sections[0].content


def test_handles_no_match_for_jsonpath_gracefully(tmp_path, _template):
    (tmp_path / "sections.toml").write_text(
        one_content_toml.replace('jsonpath = "$"', 'jsonpath = "$.nonexistent"')
    )
    config = Config(template_dir=tmp_path)

    built_sections = build_sections(properties, config)

    assert built_sections == [
        BuiltSection(output_path=Path("section1.qmd"), content=""),
    ]


def test_rejects_bad_sections_toml(tmp_path):
    (tmp_path / "sections.toml").write_text("[[one]]\noutput-path = 'no-content.qmd'")
    config = Config(template_dir=tmp_path)

    with raises(ValidationError):
        build_sections(properties, config)


def test_flags_missing_template_folder():
    config = Config(template_dir=Path("nonexistent-folder"))

    with raises(NotADirectoryError):
        build_sections(properties, config)


def test_flags_missing_sections_toml(tmp_path):
    config = Config(template_dir=tmp_path)

    with raises(FileNotFoundError):
        build_sections(properties, config)


def test_flags_missing_template_file(tmp_path):
    (tmp_path / "sections.toml").write_text(one_content_toml)
    config = Config(template_dir=tmp_path)

    with raises(FileNotFoundError):
        build_sections(properties, config)


def test_flags_bad_jsonpath_for_one_section(tmp_path, _template):
    (tmp_path / "sections.toml").write_text(
        one_content_toml.replace('jsonpath = "$"', 'jsonpath = "$.resources[*]"')
    )
    config = Config(template_dir=tmp_path)
    multi_resource_properties = {
        "name": "example-datapackage",
        "resources": [
            {
                "name": "example-resource",
                "path": "data/example-resource.csv",
            },
            {
                "name": "another-resource",
                "path": "data/another-resource.csv",
            },
        ],
    }

    with raises(ValueError):
        build_sections(multi_resource_properties, config)


# Many section ====

many_section_resources = """
[[many]]
output-path = "resources/"
content = "resources"
template-path = "resource.qmd.jinja"
jinja-variable = "resource"
"""

many_section_fields = """
[[many]]
output-path = "fields/"
content = "fields"
template-path = "field.qmd.jinja"
jinja-variable = "field"
"""

many_section_both = f"""
{many_section_resources}
{many_section_fields}
"""


def _get_resource_sections(properties: dict[str, Any]) -> list[BuiltSection]:
    return _map(
        properties["resources"],
        lambda resource: BuiltSection(
            content=resource["name"],
            output_path=Path(f"resources/{resource['name']}.qmd"),
        ),
    )


def _get_field_sections(properties: dict[str, Any]) -> list[BuiltSection]:
    return _map(
        _flat_map(
            properties["resources"], lambda resource: resource["schema"]["fields"]
        ),
        lambda field: BuiltSection(
            content=field["name"], output_path=Path(f"fields/{field['name']}.qmd")
        ),
    )


@mark.parametrize(
    "sections_toml, get_expected_sections",
    [
        (many_section_resources, lambda props: _get_resource_sections(props)),
        (many_section_fields, lambda props: _get_field_sections(props)),
        (
            many_section_both,
            lambda props: _get_resource_sections(props) + _get_field_sections(props),
        ),
    ],
)
def test_builds_many_section(
    sections_toml, get_expected_sections, tmp_path, datapackage
):
    (tmp_path / "sections.toml").write_text(sections_toml)
    (tmp_path / "resource.qmd.jinja").write_text("{{ resource.name }}")
    (tmp_path / "field.qmd.jinja").write_text("{{ field.name }}")
    config = Config(template_dir=tmp_path)

    built_sections = build_sections(datapackage, config)

    assert built_sections == get_expected_sections(datapackage)


def test_handles_no_fields_gracefully(tmp_path):
    (tmp_path / "sections.toml").write_text(many_section_fields)
    (tmp_path / "resource.qmd.jinja").write_text("{{ resource.name }}")
    (tmp_path / "field.qmd.jinja").write_text("{{ field.name }}")
    config = Config(template_dir=tmp_path)

    built_sections = build_sections(properties, config)

    assert built_sections == []


@mark.parametrize(
    ("output_path_in, output_path_out"),
    [
        ("resources/{resource-name}.qmd", "resources/example-resource.qmd"),
        ("resources/my-{resource-name}.qmd", "resources/my-example-resource.qmd"),
        ("resources/{resource-name}/index.qmd", "resources/example-resource/index.qmd"),
        (
            "resources/{resource-name}/folder",
            "resources/example-resource/folder/example-resource.qmd",
        ),
    ],
)
def test_resolves_placeholder_for_resource_files(
    tmp_path, output_path_in, output_path_out
):
    (tmp_path / "sections.toml").write_text(
        many_section_resources.replace("resources/", output_path_in)
    )
    (tmp_path / "resource.qmd.jinja").write_text("{{ resource.name }}")
    config = Config(template_dir=tmp_path)

    built_sections = build_sections(properties, config)

    assert built_sections[0].output_path == Path(output_path_out)


@mark.parametrize(
    ("output_path_in, output_path_out"),
    [
        ("fields/", "fields/example-field.qmd"),
        ("fields/{field-name}.qmd", "fields/example-field.qmd"),
        (
            "resources/{resource-name}/fields/",
            "resources/example-resource/fields/example-field.qmd",
        ),
        (
            "resources/{resource-name}/fields/{field-name}.qmd",
            "resources/example-resource/fields/example-field.qmd",
        ),
        (
            "resources/{resource-name}/{field-name}/index.qmd",
            "resources/example-resource/example-field/index.qmd",
        ),
        (
            "resources/{resource-name}/{field-name}/folder",
            "resources/example-resource/example-field/folder/example-field.qmd",
        ),
    ],
)
def test_resolves_placeholder_for_field_files(
    tmp_path, output_path_in, output_path_out
):
    (tmp_path / "sections.toml").write_text(
        many_section_fields.replace("fields/", output_path_in)
    )
    (tmp_path / "field.qmd.jinja").write_text("{{ field.name }}")
    config = Config(template_dir=tmp_path)
    one_field_properties = {
        "name": "example-datapackage",
        "resources": [
            {
                "name": "example-resource",
                "path": "data/example-resource.csv",
                "schema": {"fields": [{"name": "example-field", "type": "integer"}]},
            }
        ],
    }

    built_sections = build_sections(one_field_properties, config)

    assert built_sections[0].output_path == Path(output_path_out)


def test_builds_toml_with_one_and_many_sections(tmp_path, _template, datapackage):
    toml = f"{multi_section_toml}\n{many_section_both}"
    (tmp_path / "sections.toml").write_text(toml)
    (tmp_path / "resource.qmd.jinja").write_text("{{ resource.name }}")
    (tmp_path / "field.qmd.jinja").write_text("{{ field.name }}")
    config = Config(template_dir=tmp_path)
    package_name = datapackage["name"]

    built_sections = build_sections(datapackage, config)

    assert built_sections == [
        BuiltSection(
            output_path=Path("section1.qmd"),
            content=f"{package_name}\n{package_name}",
        ),
        BuiltSection(output_path=Path("section2.qmd"), content=package_name),
    ] + _get_resource_sections(datapackage) + _get_field_sections(datapackage)


def test_flags_missing_template_file_in_many_section(tmp_path):
    (tmp_path / "sections.toml").write_text(many_section_resources)
    config = Config(template_dir=tmp_path)

    with raises(FileNotFoundError):
        build_sections(properties, config)



"""Functions for parsing the source for a Data Package."""

from dataclasses import dataclass
from pathlib import Path
from urllib import parse


@dataclass(frozen=True)
class Address:
    """A source parsed into an actual address."""

    value: str
    local: bool


def parse_source(source: str) -> Address:
    """Parse the source of a Data Package into a formal `Address`.

    Args:
        source: The string representation for the location of a Data Package
            metadata, either as a path, `https`, or `gh`/`github` repository.

    Returns:
        A formal `Address` class.

    Raises:
        ValueError: If the `source` contains something other than what
            Flower can accept.

    Examples:
        ```{python}
        import seedcase_flower as fl
        print(fl.parse_source("./datapackage.json"))
        print(fl.parse_source("https://raw.githubusercontent.com/seedcase-project/seedcase-flower/refs/heads/main/docs/includes/datapackage.json"))
        ```
    """
    split_source = parse.urlsplit(source)
    if split_source.scheme == "":
        split_source = split_source._replace(scheme="file")
    match split_source.scheme:
        case "file":
            return _convert_to_path(split_source)
        case "https":
            return _convert_to_https(split_source)
        case "gh" | "github":
            return _convert_to_github(split_source)
        case _:
            raise ValueError(
                "The source must be either a path to an existing file or "
                "folder or have one of the following prefixes: `https:`, "
                "`gh:`, `github:`"
            )


def _convert_to_path(source: parse.SplitResult) -> Address:
    path = Path(source.path).resolve()
    if path.is_dir():
        path /= "datapackage.json"
    source = source._replace(path=path.as_posix())
    return Address(value=source.geturl(), local=True)


def _convert_to_https(source: parse.SplitResult) -> Address:
    return Address(value=source.geturl(), local=False)


def _convert_to_github(source: parse.SplitResult) -> Address:
    full_path = f"{source.netloc}{source.path}"
    if "@" in full_path:
        owner_repo, ref = full_path.rsplit("@", 1)
    else:
        owner_repo, ref = full_path, "main"

    _check_github_source(owner_repo, ref, full_path)

    return Address(
        value=source._replace(
            scheme="https",
            netloc="raw.githubusercontent.com",
            path=f"/{owner_repo}/{ref}/datapackage.json",
        ).geturl(),
        local=False,
    )


def _check_github_source(owner_repo: str, ref: str, original: str) -> None:
    parts = owner_repo.split("/")
    if len(parts) != 2 or not all(parts):
        raise ValueError(
            f"Invalid GitHub source '{original}': must be in format "
            "'gh:owner/repo', 'github:owner/repo', or with an optional "
            "'@ref' (e.g., 'gh:owner/repo@v1.0.0')."
        )

    if not ref:
        raise ValueError(
            f"Invalid GitHub source '{original}': ref after '@' cannot be empty."
        )



"""Generate the expected help-output strings used in test_cli.py.

Run this script after changing a docstring or CLI parameter. Only snippets
whose output differs from the current constants in test_cli.py are printed.
"""

import sys
from io import StringIO
from operator import itemgetter

from rich.console import Console

from seedcase_flower.cli import app
from tests.test_cli import _BUILD_HELP_PAGE, _HELP_PAGE, _VIEW_HELP_PAGE


def _capture_help(args: list[str]) -> str:
    """Return the help text produced by *args* as a plain string."""
    console = Console(
        width=90,
        force_terminal=True,
        highlight=False,
        color_system=None,
        legacy_windows=False,
    )

    old_stdout = sys.stdout
    sys.stdout = StringIO()
    try:
        try:
            app(args, console=console)
        except SystemExit:
            pass
        return sys.stdout.getvalue()
    finally:
        sys.stdout = old_stdout


def _is_outdated(check: tuple[str, list[str], str]) -> bool:
    """Return True if the current help output differs from the stored constant."""
    _, args, current = check
    return _capture_help(args) != current


def _find_outdated_checks(
    checks: list[tuple[str, list[str], str]],
) -> list[tuple[str, list[str]]]:
    """Return checks whose current help output differs from the stored constant."""
    return list(map(itemgetter(0, 1), filter(_is_outdated, checks)))


def _as_constant_snippet(name: str, text: str) -> str:
    """Return a copy-pasteable constant assignment for *text*."""
    lines = text.splitlines()
    indented_body = "\n".join(f"    {line}" if line else "" for line in lines)
    return f'{name} = dedent(\n    """\\\n{indented_body}\n    """  # noqa\n)'


if __name__ == "__main__":
    checks = [
        ("_HELP_PAGE", ["--help"], _HELP_PAGE),
        ("_BUILD_HELP_PAGE", ["build", "--help"], _BUILD_HELP_PAGE),
        ("_VIEW_HELP_PAGE", ["view", "--help"], _VIEW_HELP_PAGE),
    ]
    changed = _find_outdated_checks(checks)

    if not changed:
        print("No changes detected. All help-output constants are up to date.")
    else:
        print("\nReview that the output below looks as expected.")
        print("Then, copy and paste it into tests/test_cli.py,")
        print("replacing the variable(s) with the same name.")
        for name, args in changed:
            print("\n\n")
            print(_as_constant_snippet(name, _capture_help(args)))



from pathlib import Path
from typing import Optional

from pydantic import BaseModel

from seedcase_flower.styles import Style


class Config(BaseModel, frozen=True):
    """Configuration settings for styling the metadata.

    See the
    [design](https://flower.seedcase-project.org/docs/design/interface/config#config)
    for an explanation of how Config is designed. See the
    [guide](https://flower.seedcase-project.org/docs/guide/custom-styles) on how to
    set up custom styles and sections.
    See `Section` and `Content` help for more details on how to set up the sections.

    Attributes:
        style: The built-in style to use for outputting the
            documentation. Ignored when `template_dir` is set.
        template_dir: When using a custom style, this should be the
            relative directory path to the
            [Jinja2](https://jinja.palletsprojects.com/en/stable/) template files.
            The directory **must** contain at least one template Jinja2 file and
            a `sections.toml` file that contains attributes for the `Section` classes.
        output_dir: The directory where output files will be saved.

    Examples:
        ```{python}
        import seedcase_flower as fl
        from pathlib import Path

        # A config using the built-in `quarto-resource-listing` style and outputting
        # to the `my-docs/` folder.
        config = fl.Config(
            style=fl.Style.quarto_resource_listing, output_dir=Path("my-docs/")
        )

        # A custom style that points to a template folder and outputs
        # to the default `docs/` folder.
        config = fl.Config(template_dir=Path("templates/"))

        # A custom style that points to a template folder and outputs
        # to the `my-docs/` folder.
        config = fl.Config(template_dir=Path("templates/"), output_dir=Path("my-docs/"))
        ```
    """

    style: Style = Style.quarto_one_page
    template_dir: Optional[Path] = None
    output_dir: Path = Path("docs")



"""Custom quartodoc renderer that fixes the output of returns and raises sections."""

from __future__ import annotations

from typing import Literal, Union

from plum import dispatch
from quartodoc import MdRenderer, layout
from quartodoc._griffe_compat import docstrings as ds
from quartodoc.pandoc.blocks import DefinitionList
from quartodoc.renderers.md_renderer import ParamRow
from tabulate import tabulate


class Renderer(MdRenderer):  # type: ignore[misc]
    style = "seedcase"

    @dispatch
    def render_header(self, el: layout.Doc) -> str:
        """Render the header of a docstring, including any anchors."""
        _str_dispname = el.name

        _anchor = f"{{ #{el.obj.path} }}"

        # For lvl 1 headers, add a yml header with the ipynb-shell-interactivity setting
        # to get all output from the cell
        if self.crnt_header_level == 1:
            return f"---\nipynb-shell-interactivity: all\ntitle: {_str_dispname}\n---"
        return f"{'#' * self.crnt_header_level} {_str_dispname} {_anchor}"

    # returns ----

    @dispatch
    def render(
        self, el: Union[ds.DocstringSectionReturns, ds.DocstringSectionRaises]
    ) -> str:
        rows = list(map(self.render, el.value))
        header = [
            "Type",
            "Description",
        ]  # removed "Name" from header (only relevant when table-style is 'table')
        return self._render_table(rows, header, "returns")

    def _render_table(
        self,
        rows: list[ParamRow],
        headers: list[str],
        style: Literal["parameters", "attributes", "returns"],
    ) -> str:
        if self.table_style == "description-list":
            str_rows = str(DefinitionList([row.to_definition_list() for row in rows]))
            # remove empty fields and their separators
            return str_rows.replace(
                "<code>[:]{.parameter-annotation-sep} ", "<code>"
            ).replace("[]{.parameter-name} [:]{.parameter-annotation-sep} ", "")

        row_tuples = [row.to_tuple(style) for row in rows]
        # remove empty fields
        row_tuples_compact = [
            tuple(field for field in tup if field not in (None, ""))
            for tup in row_tuples
        ]
        return tabulate(row_tuples_compact, headers=headers, tablefmt="github")

    # Summarize ===============================================================

    @dispatch
    def summarize(self, el: layout.Section) -> str:
        desc = f"\n\n{el.desc}" if el.desc is not None else ""
        if el.title is not None:
            header = f"## {el.title}{desc}"
        elif el.subtitle is not None:
            header = f"### {el.subtitle}{desc}"
        else:
            header = ""

        if el.contents:
            thead = "| | |\n| --- | --- |"

            rendered = []
            for child in el.contents:
                rendered.append(self.summarize(child))

            str_func_table = "\n".join([thead, *rendered])
            # add colwidths
            return f"{header}\n\n{str_func_table}\n\n" + ': {tbl-colwidths="[40,60]"}'

        return header



from enum import Enum


class Style(Enum):
    """Built-in styles for generating documentation."""

    quarto_one_page = "quarto_one_page"
    quarto_resource_listing = "quarto_resource_listing"
    quarto_resource_tables = "quarto_resource_tables"


class ViewStyle(Enum):
    """Subset of styles suitable for terminal output (single-page only)."""

    quarto_one_page = "quarto_one_page"



"""Tests for the CLI commands."""

from io import StringIO
from pathlib import Path
from textwrap import dedent

import pytest
from rich.console import Console
from rich.markdown import Markdown

from seedcase_flower.build_sections import (
    BuiltSection,
    _get_template_dir,
    _load_sections_toml,
)
from seedcase_flower.cli import _CONSOLE_THEME, app
from seedcase_flower.config import Config
from seedcase_flower.parse_source import Address
from seedcase_flower.styles import Style, ViewStyle


@pytest.fixture
def mock_parse_source(mocker):
    """Mock _parse_source to isolate CLI tests from filesystem resolution."""
    return mocker.patch("seedcase_flower.cli.parse_source")


@pytest.fixture
def mock_read_properties(mocker):
    """Mock read_properties to isolate CLI tests from file I/O."""
    return mocker.patch("seedcase_flower.cli.read_properties")


@pytest.fixture
def mock_build_sections(mocker):
    """Mock build_sections to isolate CLI tests from template rendering."""
    return mocker.patch("seedcase_flower.cli.build_sections")


@pytest.fixture
def mock_write_sections(mocker):
    """Mock write_sections to isolate CLI tests from file I/O."""
    return mocker.patch("seedcase_flower.cli.write_sections")


# Testing CLI invocation ====


def test_build_with_mocked_internals(
    mock_parse_source, mock_read_properties, mock_build_sections, mock_write_sections
):
    """Isolate CLI behaviour by mocking internal helpers."""
    fake_source = Address(value="file:///datapackage.json", local=True)
    mock_parse_source.return_value = fake_source
    # Simulate running the app from the command line (but without calling sys.exit())
    app(["build", "datapackage.json"], result_action="return_value")

    # Checking that the correct values were passed to the internal functions
    mock_parse_source.assert_called_once_with("datapackage.json")
    mock_read_properties.assert_called_once_with(fake_source)
    mock_build_sections.assert_called_once_with(
        mock_read_properties.return_value, Config()
    )
    mock_write_sections.assert_called_once_with(
        mock_build_sections.return_value, Path("docs")
    )


# Checking stdout ====


def test_build_verbose_prints_output(
    capsys, datapackage_path, datapackage, tmp_path, monkeypatch
):
    """--verbose should print package name, package path, style, output dir, and created
    file paths."""
    monkeypatch.chdir(tmp_path)
    app(
        ["build", datapackage_path, "--verbose"],
        result_action="return_value",
    )

    out = capsys.readouterr().out.replace("\n", "")  # To not break long path

    for content in [
        datapackage["name"],
        datapackage_path,
        Style.quarto_one_page.name,
        "docs/",
        "docs/index.qmd",
    ]:
        assert content in out, f"Expected {content!r} to be a substring in {out!r}."


def test_build_no_verbose_produces_no_output(
    capsys, datapackage_path, tmp_path, monkeypatch
):
    """Without --verbose, build should produce no stdout."""
    monkeypatch.chdir(tmp_path)
    app(["build", datapackage_path], result_action="return_value")
    assert capsys.readouterr().out == ""


# File-based config ====


def test_build_reads_source_from_flower_toml(tmp_path, monkeypatch):
    """Build args specified in .flower.toml should overwrite the default values."""
    toml_path = tmp_path / ".flower.toml"
    toml_path.write_text(
        'source = "custom.json"\n'
        'style = "quarto_resource_listing"\n'
        'template_dir = "my-templates/"\n'
        'output_dir = "my-docs/"\n'
        "verbose = true\n"
    )

    monkeypatch.chdir(tmp_path)

    _, bound, _ = app.parse_args(["build"])
    assert bound.arguments["source"] == "custom.json"
    assert bound.arguments["style"] == Style.quarto_resource_listing
    assert bound.arguments["template_dir"] == Path("my-templates/")
    assert bound.arguments["output_dir"] == Path("my-docs/")
    assert bound.arguments["verbose"] is True


def test_view_ignores_flower_toml(tmp_path, monkeypatch):
    """View should ignore any .flower.toml config and always use defaults."""
    toml_path = tmp_path / ".flower.toml"
    toml_path.write_text(
        'source = "custom.json\n'  # a comment to prevent ruff from wrapping this
        'style = "quarto_one_page"\n'
    )
    monkeypatch.chdir(tmp_path)

    _, bound, _ = app.parse_args(["view"])
    assert "source" not in bound.arguments
    assert "style" not in bound.arguments


# Help output ====

_HELP_PAGE = dedent(
    """\
    Usage: seedcase-flower COMMAND

    Flower generates human-readable documentation from Data Packages.

    ╭─ Commands ─────────────────────────────────────────────────────────────────────────────╮
    │ <build>               Build human-readable documentation from a datapackage.json file. │
    │ <view>                Display the contents of a datapackage.json in a human-friendly   │
    │                       way.                                                             │
    │ --help                Display this message and exit.                                   │
    │ --install-completion  Install shell completion for this application.                   │
    │ --version             Display application version.                                     │
    ╰────────────────────────────────────────────────────────────────────────────────────────╯
    """  # noqa
)


_BUILD_HELP_PAGE = dedent(
    """\
    Usage: seedcase-flower build [ARGS]

    Build human-readable documentation from a datapackage.json file.

    ╭─ Parameters ───────────────────────────────────────────────────────────────────────────╮
    │ --source <SOURCE>              The location of a datapackage.json, defaults to a file  │
    │                                or folder path. Can also be an https: source to a       │
    │                                remote datapackage.json or a github: / gh: pointing to  │
    │                                a repo with a datapackage.json in the repo root (in the │
    │                                format gh:org/repo, which can also include reference to │
    │                                a tag or branch, such as gh:org/repo@main or            │
    │                                gh:org/repo@1.0.1).                                     │
    │                                [default: datapackage.json]                             │
    │ --style <STYLE>                The style used to structure the output. If a template   │
    │                                directory is given, this parameter will be ignored.     │
    │                                [choices: quarto-one-page, quarto-resource-listing,     │
    │                                quarto-resource-tables]                                 │
    │                                [default: quarto-one-page]                              │
    │ --template-dir <TEMPLATE-DIR>  The directory that contains the Jinja template files    │
    │                                and sections.toml. When set, it will override any       │
    │                                built-in style given by the style parameter.            │
    │                                [default: None]                                         │
    │ --output-dir <OUTPUT-DIR>      The directory to save the generated files in.           │
    │                                [default: docs]                                         │
    │ --verbose                      If True, prints additional information to the console.  │
    │                                [default: False]                                        │
    ╰────────────────────────────────────────────────────────────────────────────────────────╯
    """  # noqa
)

_VIEW_HELP_PAGE = dedent(
    """\
    Usage: seedcase-flower view [ARGS]

    Display the contents of a datapackage.json in a human-friendly way.

    ╭─ Parameters ───────────────────────────────────────────────────────────────────────────╮
    │ --source <SOURCE>  The location of a datapackage.json, defaults to a file or folder    │
    │                    path. Can also be an https: source to a remote datapackage.json or  │
    │                    a github: / gh: pointing to a repo with a datapackage.json in the   │
    │                    repo root (in the format gh:org/repo, which can also include        │
    │                    reference to a tag or branch, such as gh:org/repo@main or           │
    │                    gh:org/repo@1.0.1).                                                 │
    │                    [default: datapackage.json]                                         │
    │ --style <STYLE>    The style used to display the output in the terminal. Must be a     │
    │                    single-page style.                                                  │
    │                    [choices: quarto-one-page]                                          │
    │                    [default: quarto-one-page]                                          │
    ╰────────────────────────────────────────────────────────────────────────────────────────╯
    """  # noqa
)

_CHANGED_MSG = (
    "The `{cmd}` help output changed. Run `just generate-help-strings` "
    "and paste the updated string into the relevant test."
)


@pytest.fixture
def console():
    return Console(
        width=90,
        force_terminal=True,
        highlight=False,
        color_system=None,
        legacy_windows=False,
    )


def test_help_page(capsys, console):
    """Top-level --help should match expected output."""
    with pytest.raises(SystemExit):
        app(["--help"], console=console)
    assert capsys.readouterr().out == _HELP_PAGE, _CHANGED_MSG.format(cmd="general")


def test_build_help_page(capsys, console):
    """build --help should document all parameters with defaults and choices."""
    with pytest.raises(SystemExit):
        app(["build", "--help"], console=console)
    assert capsys.readouterr().out == _BUILD_HELP_PAGE, _CHANGED_MSG.format(cmd="build")


# It was not possible to include these color markup tags directly in the help string
# test above because printing them out explicitly in the rich console messes up the
# column widths in cyclopts
def test_build_help_page_applies_rich_markup(capsys):
    """build --help should apply bold-cyan to flags and dim to placeholders."""
    markup_console = Console(
        width=90,
        force_terminal=False,
        highlight=False,
        color_system=None,
        markup=False,
        legacy_windows=False,
    )
    with pytest.raises(SystemExit):
        app(["build", "--help"], console=markup_console)
    output = capsys.readouterr().out
    assert "[bold cyan]--source[/bold cyan]" in output
    assert "[dim]<SOURCE>[/dim]" in output
    assert "[bold cyan]--style[/bold cyan]" in output
    assert "[dim]<STYLE>[/dim]" in output
    assert "[bold cyan]--verbose[/bold cyan]" in output
    # Boolean flags must not produce a positional placeholder
    assert "[dim]<verbose>[/dim]" not in output


# view ====


def test_view_styles_are_one_page():
    """Every ViewStyle member must map to a single-section (one-page) style."""
    for member in ViewStyle:
        style = Style[member.name]
        sections = _load_sections_toml(_get_template_dir(style))
        assert not sections.many_sections, (
            f"ViewStyle.{member.name} includes `Many` sections, "
            "but view styles must be single-page (exactly 1 `One` section)."
        )
        assert len(sections.one_sections) == 1, (
            f"ViewStyle.{member.name} has {len(sections.one_sections)} sections, "
            "but view styles must be single-page (exactly 1 `One` section)."
        )


def test_view_with_mocked_internals(mocker):
    """view should parse source, build sections, and render via Console."""
    mock_parse_source = mocker.patch("seedcase_flower.cli.parse_source")
    mock_read_properties = mocker.patch("seedcase_flower.cli.read_properties")
    mock_build_sections = mocker.patch("seedcase_flower.cli.build_sections")
    mock_console_cls = mocker.patch("seedcase_flower.cli.Console")
    mock_console = mock_console_cls.return_value

    mock_build_sections.return_value = [
        BuiltSection(content="# Test", output_path=None)
    ]

    fake_source = Address(value="file:///datapackage.json", local=True)
    mock_parse_source.return_value = fake_source

    app(["view", "datapackage.json"], result_action="return_value")

    mock_parse_source.assert_called_once_with("datapackage.json")
    mock_read_properties.assert_called_once_with(fake_source)
    mock_build_sections.assert_called_once_with(
        mock_read_properties.return_value,
        Config(style=Style.quarto_one_page),
    )
    assert mock_console.print.called


# The color codes cannot easily be tested so we look at substrings
# instead of the exact full output
def test_view_renders_datapackage(capsys, datapackage_path):
    """view should render all key datapackage fields to the terminal."""
    app(["view", datapackage_path], result_action="return_value")
    output = capsys.readouterr().out

    # Package metadata
    assert "test-package" in output
    assert "Test Package" in output
    assert "MIT" in output
    assert "1.0.0" in output
    assert "A test datapackage" in output

    # Resource structure
    assert "Resources" in output
    assert "data" in output
    assert "data.csv" in output

    # Schema fields table
    assert "id" in output
    assert "integer" in output
    assert "name" in output
    assert "string" in output


def test_styled_markdown_table_renders_box_and_header():
    """Markdown tables should render with a heavy-head box and column separators."""
    md = "| A | B |\n|---|---|\n| 1 | 2 |\n"
    out = StringIO()
    Console(file=out, theme=_CONSOLE_THEME, no_color=True).print(Markdown(md))
    output = out.getvalue()
    assert "┏" in output  # heavy outer box top
    assert "┡" in output  # heavy-to-light header separator
    assert "┴" in output  # bottom border with column joins
    assert "A" in output
    assert "B" in output


def test_view_help_page(capsys, console):
    """view --help should document source and style parameters."""
    with pytest.raises(SystemExit):
        app(["view", "--help"], console=console)
    assert capsys.readouterr().out == _VIEW_HELP_PAGE, _CHANGED_MSG.format(cmd="view")



from pathlib import Path

from pytest import mark, raises

from seedcase_flower.sections import Content, Many, ManyContent, One

# One section ====

content = Content(
    jsonpath="$.resources",
    template_path=Path("resource_template.html.jinja"),
    jinja_variable="resource",
)


def test_creates_one_section():
    section = One(
        output_path=Path("resources/"),
        contents=[content],
    )
    assert section.output_path == Path("resources/")
    assert section.contents == [content]

    section = One(
        contents=[content] * 2,
    )
    assert section.output_path is None
    assert section.contents == [content] * 2


def test_cannot_create_section_with_absolute_output_path():
    with raises(ValueError):
        One(
            output_path=Path("/absolute/path/resources/"),
            contents=[content],
        )


def test_cannot_create_content_with_bad_jsonpath():
    with raises(ValueError):
        Content(
            jsonpath="<><>bad.path",
            template_path=Path("template.qmd.jinja"),
            jinja_variable="data",
        )


def test_cannot_create_content_with_absolute_template_path():
    with raises(ValueError):
        Content(
            jsonpath="$",
            template_path=Path("/absolute/path/template.qmd.jinja"),
            jinja_variable="data",
        )


@mark.parametrize("template_path", ["", "template.qmd", "template.jinja", "template"])
def test_cannot_create_content_without_extension_in_template_path(template_path):
    with raises(ValueError):
        Content(
            jsonpath="$",
            template_path=Path(template_path),
            jinja_variable="data",
        )


# Many section ====


@mark.parametrize(
    "output_path_in, output_path_out",
    [
        (None, None),
        (Path("resources/"), Path("resources/{resource-name}.qmd")),
        (Path("resources/{resource-name}.qmd"), Path("resources/{resource-name}.qmd")),
        (
            Path("resources/{resource-name}/index.qmd"),
            Path("resources/{resource-name}/index.qmd"),
        ),
        (
            Path("resources/{resource-name}/folder"),
            Path("resources/{resource-name}/folder/{resource-name}.qmd"),
        ),
    ],
)
def test_creates_many_section_for_resources(output_path_in, output_path_out):
    section = Many(
        output_path=output_path_in,
        content=ManyContent.resources,
        template_path=Path("resource.qmd.jinja"),
        jinja_variable="resource",
    )
    assert section.output_path == output_path_out


@mark.parametrize(
    "output_path_in, output_path_out",
    [
        (None, None),
        (Path("fields/"), Path("fields/{field-name}.qmd")),
        (Path("fields/{field-name}.qmd"), Path("fields/{field-name}.qmd")),
        (
            Path("resources/{resource-name}/fields/"),
            Path("resources/{resource-name}/fields/{field-name}.qmd"),
        ),
        (
            Path("resources/{resource-name}/fields/{field-name}.qmd"),
            Path("resources/{resource-name}/fields/{field-name}.qmd"),
        ),
        (
            Path("resources/{resource-name}/{field-name}/index.qmd"),
            Path("resources/{resource-name}/{field-name}/index.qmd"),
        ),
        (
            Path("resources/{resource-name}/{field-name}/folder"),
            Path("resources/{resource-name}/{field-name}/folder/{field-name}.qmd"),
        ),
    ],
)
def test_creates_many_section_for_fields(output_path_in, output_path_out):
    section = Many(
        output_path=output_path_in,
        content=ManyContent.fields,
        template_path=Path("field.qmd.jinja"),
        jinja_variable="field",
    )
    assert section.output_path == output_path_out


@mark.parametrize(
    "many_content, output_path",
    [
        (ManyContent.resources, "resources/{not-resource-name}.qmd"),
        (ManyContent.resources, "resources/{field-name}.qmd"),
        (ManyContent.resources, "{resource-name}/{resource-name}.qmd"),
        (ManyContent.resources, "points/to/file/index.qmd"),
        (ManyContent.fields, "fields/{not-field-name}.qmd"),
        (ManyContent.fields, "{field-name}/{resource-name}/index.qmd"),
        (ManyContent.fields, "{field-name}/{field-name}/index.qmd"),
        (ManyContent.fields, "{resource-name}/{resource-name}/index.qmd"),
        (ManyContent.fields, "points/to/file/index.qmd"),
        (ManyContent.fields, "resources/{resource-name}.qmd"),
        (ManyContent.fields, "resources/{resource-name}/index.qmd"),
    ],
)
def test_cannot_create_many_section_with_bad_output_path(many_content, output_path):
    with raises(ValueError):
        Many(
            output_path=output_path,
            content=many_content,
            template_path=Path("template.qmd.jinja"),
            jinja_variable="field",
        )



# ruff: noqa
# mypy: ignore-errors
jsonpath  # unused variable (src/seedcase_flower/section.py:78)
template_path  # unused variable (src/seedcase_flower/section.py:79)
jinja_variable  # unused variable (src/seedcase_flower/section.py:80)
mode  # unused variable (src/seedcase_flower/section.py:81)
_check_jsonpath  # unused method (src/seedcase_flower/section.py:83)
quarto_resource_tables  # unused variable (src/seedcase_flower/styles.py:9)
cls  # unused variable (src/seedcase_flower/section.py:85)
target  # unused variable (src/seedcase_flower/.venv/lib/python3.12/site-packages/_virtualenv.py:50)
handler  # unused variable (src/seedcase_flower/internals.py:20)
source  # unused variable (src/seedcase_flower/internals.py:20)



"""Helper functions for internal use."""

from itertools import chain, repeat
from typing import Any, Callable, Iterable, Optional, TypeVar

from cyclopts.annotations import get_hint_name
from cyclopts.help import HelpEntry
from rich import print as rprint


def _format_param_help(entry: HelpEntry) -> str:
    """Re-structure the parameter help into a more readable format."""
    # Sort to put the flag first (eg `--source SOURCE` instead of the default
    # `SOURCE --source`)
    names = map(_add_highlight_syntax, sorted(entry.names), repeat(entry.type))
    return f"{' '.join(names)}".strip()


def _add_highlight_syntax(name: str, entry_type: Optional[type]) -> str:
    """Add markup character to highlight in colors, etc where desired."""
    formatted_name = f"[bold cyan]{name}[/bold cyan]"
    if not name.startswith("-"):
        # Matching the `dim` used by default in cyclopts for `choices` and
        # `defaults` in the description
        formatted_name = f"[dim]<{name}>[/dim]"

        # Don't output redundant value placeholder for boolean flags
        if get_hint_name(entry_type) == "bool":
            formatted_name = ""
    return formatted_name


In = TypeVar("In")
Out = TypeVar("Out")


def _map(x: Iterable[In], fn: Callable[[In], Out]) -> list[Out]:
    return list(map(fn, x))


def _filter(x: Iterable[In], fn: Callable[[In], bool]) -> list[In]:
    return list(filter(fn, x))


def _flat_map(items: Iterable[In], fn: Callable[[In], Iterable[Out]]) -> list[Out]:
    """Maps and flattens the items by one level."""
    return list(chain.from_iterable(map(fn, items)))


def _pretty_print(verbose: bool, message: str) -> None:
    if verbose:
        rprint(message)


def _number(value: str, items: list[Any]) -> str:
    suffix = "" if len(items) == 1 else "s"
    return f"{len(items)} {value}{suffix}"



"""Custom exception handling for seedcase-flower."""


class FlowerError(Exception):
    """Base class for all seedcase-flower exceptions."""

    def __init__(self, resolved_address: str, reason: str = "") -> None:
        """Initialize FlowerError with resolved_address and optional reason."""
        message = f"Could not load '{resolved_address}'.\n"
        if reason:
            message += f"{reason}"
        super().__init__(message)


class FileDoesNotExistError(FlowerError):
    """Error when a local path does not point to an existing file."""

    def __init__(self, resolved_address: str) -> None:
        """Initialize FileDoesNotExistError with resolved_address."""
        super().__init__(
            resolved_address,
            "The file cannot be found; maybe there is a typo in the path?",
        )


class JSONFormatError(FlowerError):
    """Error when a file has invalid JSON format."""

    def __init__(self, resolved_address: str, json_error: str) -> None:
        """Initialize JSONFormatError with resolved_address and JSON error details."""
        super().__init__(
            resolved_address,
            f"A JSON formatting issue was found: {json_error}",
        )


class HTTPStatusError(FlowerError):
    """Error when an HTTP request returns an error status code."""

    def __init__(self, resolved_address: str, code: int, reason: str) -> None:
        """Initialize HTTPStatusError with resolved_address, status code, and reason."""
        super().__init__(resolved_address, f"Error code {code}: {reason}")


class HTTPDomainError(FlowerError):
    """Error when unable to connect to server due to domain not being found."""

    def __init__(self, resolved_address: str) -> None:
        """Initialize HTTPDomainError with resolved_address."""
        super().__init__(
            resolved_address,
            "Couldn't connect to the server because the domain wasn't found.",
        )


class NotJSONError(FlowerError):
    """Error when a URL does not return JSON content."""

    def __init__(self, resolved_address: str, content_type: str) -> None:
        """Initialize NotJSONError with resolved_address and actual content type."""
        super().__init__(
            resolved_address,
            f"Expected JSON but received '{content_type}'.",
        )



"""Functions for the exposed CLI."""

from pathlib import Path
from typing import Any, Optional

from cyclopts import App, Parameter, config
from cyclopts.help import ColumnSpec, DefaultFormatter, DescriptionRenderer
from rich import box
from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text
from rich.theme import Theme

from seedcase_flower.build_sections import build_sections
from seedcase_flower.config import Config
from seedcase_flower.internals import (
    _format_param_help,
    _map,
    _number,
    _pretty_print,
)
from seedcase_flower.parse_source import Address, parse_source
from seedcase_flower.read_properties import read_properties
from seedcase_flower.styles import Style, ViewStyle
from seedcase_flower.write_sections import write_sections

# To style markdown tables with a box (pipes) surrounding each column
# instead of only a horizontal line after the table header
box.SIMPLE = box.HEAVY_HEAD

_CONSOLE_THEME = Theme(
    {
        "markdown.h1": "bold yellow",
        "markdown.h2": "bold yellow",
        "markdown.h3": "italic yellow",
        "markdown.code": "blue",
        "markdown.link": "underline cyan",
        "markdown.link_url": "underline cyan",
        "markdown.table.header": "yellow",
        "markdown.table.border": "white",
    }
)

app = App(
    name="seedcase-flower",
    help="Flower generates human-readable documentation from Data Packages.",
    help_formatter=DefaultFormatter(
        column_specs=(
            ColumnSpec(renderer=_format_param_help),
            ColumnSpec(renderer=DescriptionRenderer(newline_metadata=True)),
        )
    ),
    default_parameter=Parameter(negative=(), show_default=True),
    config=[
        config.Toml(
            ".flower.toml",
            search_parents=True,
            use_commands_as_keys=False,
        ),
        config.Toml(
            "pyproject.toml",
            root_keys=["tool", "seedcase-flower"],
            search_parents=True,
            use_commands_as_keys=False,
        ),
    ],
)
app.register_install_completion_command()


@app.command()
def build(
    source: str = "datapackage.json",
    style: Style = Style.quarto_one_page,
    template_dir: Optional[Path] = None,
    output_dir: Path = Path("docs"),
    verbose: bool = False,
) -> None:
    """Build human-readable documentation from a `datapackage.json` file.

    Args:
        source: The location of a `datapackage.json`, defaults to a file or folder
            path. Can also be an `https:` source to a remote `datapackage.json` or a
            `github:` / `gh:` pointing to a repo with a `datapackage.json`
            in the repo root (in the format `gh:org/repo`, which can also include
            reference to a tag or branch, such as `gh:org/repo@main` or
            `gh:org/repo@1.0.1`).
        style: The style used to structure the output. If a template directory
            is given, this parameter will be ignored.
        template_dir: The directory that contains the Jinja template
            files and `sections.toml`. When set, it will override any
            built-in style given by the `style` parameter.
        output_dir: The directory to save the generated files in.
        verbose: If True, prints additional information to the console.
    """
    config = Config(
        style=style,
        template_dir=template_dir,
        output_dir=output_dir,
    )
    address: Address = parse_source(source)
    properties: dict[str, Any] = read_properties(address)
    _pretty_print(
        verbose, f"Read Data Package {properties['name']!r} from {address.value!r}."
    )

    built_sections = build_sections(properties, config)
    _pretty_print(
        verbose,
        (
            f"Created {_number('section', built_sections)} "
            f"using the {style.name!r} style."
        ),
    )

    output_files: list[Path] = write_sections(built_sections, output_dir)
    _pretty_print(
        verbose, f"Created {_number('file', output_files)} in '{output_dir}/':"
    )
    _pretty_print(verbose, "\n".join(_map(output_files, lambda file: f"  - '{file}'")))


@app.command(config=[])
def view(
    source: str = "datapackage.json",
    style: ViewStyle = ViewStyle.quarto_one_page,
) -> None:
    """Display the contents of a `datapackage.json` in a human-friendly way.

    Args:
        source: The location of a `datapackage.json`, defaults to a file or folder
            path. Can also be an `https:` source to a remote `datapackage.json` or a
            `github:` / `gh:` pointing to a repo with a `datapackage.json`
            in the repo root (in the format `gh:org/repo`, which can also include
            reference to a tag or branch, such as `gh:org/repo@main` or
            `gh:org/repo@1.0.1`).
        style: The style used to display the output in the terminal. Must be a
            single-page style.
    """
    address: Address = parse_source(source)
    properties: dict[str, Any] = read_properties(address)
    built_sections = build_sections(properties, Config(style=Style[style.name]))
    console = Console(theme=_CONSOLE_THEME)
    print()  # One line separation between the command and the datapackage title
    console.print(Markdown(built_sections[0].content))


def _pretty_print_error(e: Exception) -> None:
    console = Console(stderr=True)
    text = Text.from_markup(str(e))
    # Make `text` appear as if it was printed by rich.print
    pretty_text = ReprHighlighter()(text)
    console.print(
        Panel(
            pretty_text,
            title=type(e).__name__,
            border_style="red",
            box=box.ROUNDED,
            title_align="left",
        )
    )


def main() -> None:
    """Suppress traceback when running from CLI."""
    try:
        app()
    except Exception as e:
        _pretty_print_error(e)
        raise SystemExit(1)



"""Tests for the read_properties function."""

import json
from email.message import Message
from pathlib import Path
from urllib.error import HTTPError, URLError

import pytest
from check_datapackage.check import DataPackageError

from seedcase_flower.errors import (
    FileDoesNotExistError,
    HTTPDomainError,
    HTTPStatusError,
    JSONFormatError,
    NotJSONError,
)
from seedcase_flower.parse_source import Address, parse_source
from seedcase_flower.read_properties import read_properties

# read_properties: local file ====


def test_read_properties_local_filepath(datapackage_path, datapackage):
    """Reading a local datapackage.json file should return its contents."""
    address = Address(value=str(datapackage_path), local=True)
    result = read_properties(address)

    assert result == datapackage


def test_read_properties_local_dirpath(datapackage_path, datapackage):
    """Passing a path to a directory containing a datapackage.json should work."""
    address = parse_source(str(Path(datapackage_path).parent))
    result = read_properties(address)

    assert result == datapackage


def test_read_properties_raises_on_invalid_datapackage(tmp_path):
    """An invalid datapackage should raise a ValueError."""
    invalid_datapackage = {"name": "invalid-package", "resources": []}
    json_file = tmp_path / "datapackage.json"
    json_file.write_text(json.dumps(invalid_datapackage))

    address = Address(value=str(json_file), local=True)

    with pytest.raises(DataPackageError, match="should be non-empty"):
        read_properties(address)


def test_read_properties_raises_on_file_not_found():
    """A non-existent file should raise FileDoesNotExistError."""
    address = Address(value="file:///nonexistent/path/datapackage.json", local=True)

    with pytest.raises(FileDoesNotExistError):
        read_properties(address)


def test_read_properties_raises_on_malformed_json(tmp_path):
    """A file with malformed JSON should raise JSONFormatError."""
    json_file = tmp_path / "datapackage.json"
    json_file.write_text("{ invalid json }")

    address = Address(value=str(json_file), local=True)

    with pytest.raises(JSONFormatError):
        read_properties(address)


# read_properties: remote file ====


@pytest.mark.usefixtures("mocker")
def test_read_properties_remote_url(mocker, datapackage):
    """Reading a remote datapackage.json URL should return its contents."""
    mock_urlopen = mocker.patch("seedcase_flower.read_properties.request.urlopen")
    mock_response = mock_urlopen.return_value.__enter__.return_value
    mock_response.read.return_value = json.dumps(datapackage).encode()

    address = Address(value="https://example.com/datapackage.json", local=False)
    result = read_properties(address)

    assert result == datapackage
    mock_urlopen.assert_called_once_with("https://example.com/datapackage.json")


@pytest.mark.usefixtures("mocker")
def test_read_properties_raises_on_remote_invalid_json(mocker):
    """Remote URL returning invalid JSON should raise JSONFormatError."""
    mock_urlopen = mocker.patch("seedcase_flower.read_properties.request.urlopen")
    mock_response = mock_urlopen.return_value.__enter__.return_value
    mock_response.read.return_value = b"{ invalid json }"

    address = Address(value="https://example.com/datapackage.json", local=False)

    with pytest.raises(JSONFormatError):
        read_properties(address)


@pytest.mark.usefixtures("mocker")
def test_read_properties_raises_on_remote_404(mocker):
    """A remote URL returning 404 should raise HTTPStatusError."""
    mocker.patch(
        "seedcase_flower.read_properties.request.urlopen",
        side_effect=HTTPError(
            "https://example.com/datapackage.json", 404, "Not Found", Message(), None
        ),
    )

    address = Address(value="https://example.com/datapackage.json", local=False)

    with pytest.raises(HTTPStatusError):
        read_properties(address)


@pytest.mark.usefixtures("mocker")
def test_read_properties_raises_on_remote_dns_failure(mocker):
    """Remote URL with invalid domain should raise HTTPDomainError."""
    mocker.patch(
        "seedcase_flower.read_properties.request.urlopen",
        side_effect=URLError(reason=Exception("[Errno -2] Name or service not known")),
    )

    address = Address(
        value="https://nonexistent-domain-12345.com/datapackage.json", local=False
    )

    with pytest.raises(HTTPDomainError):
        read_properties(address)


@pytest.mark.usefixtures("mocker")
def test_read_properties_raises_on_non_json_content_type(mocker):
    """Remote URL returning non-JSON content type should raise NotJSONError."""
    mock_urlopen = mocker.patch("seedcase_flower.read_properties.request.urlopen")
    mock_response = mock_urlopen.return_value.__enter__.return_value
    mock_response.headers.get.return_value = "text/html; charset=utf-8"

    address = Address(value="https://example.com/datapackage.json", local=False)

    with pytest.raises(NotJSONError, match="Expected JSON but received 'text/html"):
        read_properties(address)
